@@ -2,7 +2,7 @@
 #
 # Python functions to read, split and apply patches.
 #
-# Copyright (C) 2014-2016 Sebastian Lackner
+# Copyright (C) 2014-2017 Sebastian Lackner
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -27,6 +27,7 @@ import itertools
 import os
 import re
 import shutil
+import stat
 import subprocess
 import sys
 import tempfile
@@ -135,6 +136,16 @@ class _PatchReader(object):
         tmp, self.peeked = self.peeked, None
         return tmp[1]
 
+    def read_multiline(self):
+        """Read multiline data from a patch file."""
+        line = self.read().rstrip("\r\n")
+        while True:
+            tmp = self.peek()
+            if not tmp.startswith(" "): break
+            line += tmp.rstrip("\r\n")
+            self.read()
+        return line
+
     def read_hunk(self):
         """Read one hunk from a patch file."""
         line = self.peek()
@@ -341,18 +352,12 @@ def read_patch(filename, fp=None):
                 break
 
             elif line.startswith("From: "):
-                header['author'], header['email'] = _parse_author(line[6:])
+                author = fp.read_multiline()[6:]
+                header['author'], header['email'] = _parse_author(author)
                 header.pop('signedoffby', None)
-                fp.read()
 
             elif line.startswith("Subject: "):
-                subject = line[9:].rstrip("\r\n")
-                fp.read()
-                while True:
-                    line = fp.peek()
-                    if not line.startswith(" "): break
-                    subject += line.rstrip("\r\n")
-                    fp.read()
+                subject = fp.read_multiline()[9:]
                 subject, revision = _parse_subject(subject)
                 if not subject.endswith("."): subject += "."
                 subject = re.sub('^([^:]*: *)([a-z])', lambda x: "%s%s" %
@@ -501,7 +506,7 @@ def generate_ifdef_patch(original, patched, ifdef):
     #
 
     with tempfile.NamedTemporaryFile(mode='w+') as diff:
-        exitcode = subprocess.call(["diff", "-U", "1", original.name, patched.name],
+        exitcode = subprocess.call(["git", "diff", "--no-index", "--minimal", "-U1", original.name, patched.name],
                                    stdout=diff, stderr=_devnull)
         if exitcode == 0:
             return None
@@ -516,7 +521,11 @@ def generate_ifdef_patch(original, patched, ifdef):
         fp = _PatchReader(diff.name, diff)
         fp.seek(0)
 
-        # We expect this output format from 'diff', if this is not the case things might go wrong.
+        # We expect this output format from 'git diff', if this is not the case things might go wrong.
+        line = fp.read()
+        assert line.startswith("diff --git ")
+        line = fp.read()
+        assert line.startswith("index ")
         line = fp.read()
         assert line.startswith("--- ")
         line = fp.read()
@@ -640,7 +649,7 @@ def generate_ifdef_patch(original, patched, ifdef):
 
         # Now we can finally compute the diff between the original file and our intermediate file
         diff = tempfile.NamedTemporaryFile(mode='w+')
-        exitcode = subprocess.call(["git", "diff", "--no-index", original.name, intermediate.name],
+        exitcode = subprocess.call(["git", "diff", "--no-index", "--minimal", original.name, intermediate.name],
                                    stdout=diff, stderr=_devnull)
         if exitcode != 1: # exitcode 0 cannot (=shouldn't) happen in this situation
             raise PatchDiffError("Failed to compute diff (exitcode %d)." % exitcode)
@@ -661,6 +670,14 @@ def generate_ifdef_patch(original, patched, ifdef):
     # Return the final diff
     return diff
 
+def escape_sh(text):
+    """Escape string for usage in shell '...' quotes."""
+    return text.replace("'", "'\\''")
+
+def escape_c(text):
+    """Escape string for usage in C "..." quotes."""
+    return text.replace("\\", "\\\\").replace("\"", "\\\"")
+
 if __name__ == "__main__":
     import unittest
 
@@ -944,4 +961,65 @@ if __name__ == "__main__":
             lines = diff.read().rstrip("\n").split("\n")
             self.assertEqual(lines, expected)
 
+    # Basic tests for escape_sh()
+    class EscapeShellTests(unittest.TestCase):
+        ascii = "".join([chr(i) for i in range(32, 127)] + \
+                        ["\\" + chr(i) for i in range(32, 127)])
+
+        def test_bash(self):
+            script = tempfile.NamedTemporaryFile(mode='w+', delete=False)
+            try:
+                script.write("#!/bin/bash\nprintf '%%s\\n' '%s'\n" % escape_sh(self.ascii))
+                script.close()
+
+                st = os.stat(script.name)
+                os.chmod(script.name, st.st_mode | stat.S_IEXEC)
+
+                result = subprocess.check_output([script.name])
+                if not isinstance(result, str): result = result.decode('utf-8')
+                self.assertEqual(result.rstrip("\n"), self.ascii)
+            finally:
+                os.unlink(script.name)
+
+        @unittest.skipUnless(os.path.exists("/bin/dash"), "requires Dash")
+        def test_dash(self):
+            script = tempfile.NamedTemporaryFile(mode='w+', delete=False)
+            try:
+                script.write("#!/bin/dash\nprintf '%%s\\n' '%s'\n" % escape_sh(self.ascii))
+                script.close()
+
+                st = os.stat(script.name)
+                os.chmod(script.name, st.st_mode | stat.S_IEXEC)
+
+                result = subprocess.check_output([script.name])
+                if not isinstance(result, str): result = result.decode('utf-8')
+                self.assertEqual(result.rstrip("\n"), self.ascii)
+            finally:
+                os.unlink(script.name)
+
+    # Basic tests for escape_c()
+    class EscapeCTests(unittest.TestCase):
+        ascii = "".join([chr(i) for i in range(32, 127)] + \
+                        ["\\" + chr(i) for i in range(32, 127)])
+
+        def test_c(self):
+            source = tempfile.NamedTemporaryFile(mode='w+', suffix='.c', delete=False)
+            try:
+                source.write("#include <stdio.h>\n")
+                source.write("int main() { printf(\"%%s\", \"%s\"); return 0; }\n" % escape_c(self.ascii))
+                source.close()
+
+                compiled = tempfile.NamedTemporaryFile(mode='w+', delete=False)
+                try:
+                    compiled.close()
+                    subprocess.call(["gcc", source.name, "-o", compiled.name])
+
+                    result = subprocess.check_output([compiled.name])
+                    if not isinstance(result, str): result = result.decode('utf-8')
+                    self.assertEqual(result.rstrip("\n"), self.ascii)
+                finally:
+                    os.unlink(compiled.name)
+            finally:
+                os.unlink(source.name)
+
     unittest.main()
